/// Lightweight object Serialization and Deserialization (SerDes).
///
/// In short, a "pickle" represents a [Memento](https://en.wikipedia.org/wiki/Memento_pattern)
/// (design pattern). Pickles can be passed around, persisted on disk or in the
/// cloud, and used to reproduce the original object.
///
/// This library aims to be lightweight and shies away from reflection magic or
/// intrusive changes to your classes. In fact, you can probably use the system
/// without making any changes to your serializable classes, though you will
/// need to map the attributes that define your class to the [Pickle] attributes.
///
/// ```dart
/// Pickle picklePerson(final Person person) {
///   return Pickle()
///     .withString('name', Person.name),
///     .withDouble('age', Person.age);
/// }
///
/// Person depicklePerson(final Pickle pickle) {
///   return Person(
///     name: pickle.readString('name'),
///     age: pickle.readDouble('age'),
///   );
/// }
/// ```
///
/// For full integration, have your serializable class implement [Pickleable].
/// This will mean the serializable class gets a "depickle" method and gives you
/// access to some handy SerDes methods on the [Pickle] class itself.
///
/// ```dart
/// class Person implements Pickleable {...}
///
/// final pickle = Pickle()
///   .withPickleable('mom', Person('Nancy', 34.6))
///   .withPickleable('dad', Person('Bob', 32.3))
///   .withPickleables('children', [
///       Person('harry', 4.5),
///       Person('margaret', 3.2),
///     ]);
/// ```
library fling_pickle;

// imports
import 'dart:convert';

// core
part 'src/pickler.dart';
part 'src/pickle.dart';

// picklers
part 'src/binary_pickler.dart';
part 'src/json_pickler.dart';
