import 'package:fling_pickle/fling_pickle.dart';
import 'package:test/test.dart';

class Tester implements Pickleable {
  final String name;

  Tester({final this.name = 'Fred'});

  static Tester fromPickle(final Pickle pickle) =>
      Tester(name: pickle.readString('name'));

  @override
  Pickle asPickle() {
    return PickleBuilder().withString('name', name).build();
  }

  @override
  bool operator ==(final dynamic other) =>
      other is Tester && other.name == name;
}

void main() {
  group('withBool', () {
    test('assigned and read', () {
      // given
      final pickle = PickleBuilder().withBool('key', true).build();

      // when
      final result = pickle.readBool('key');

      // then
      expect(result, true);
    });
    test('without assignment', () {
      // given
      final pickle = PickleBuilder().build();

      // then
      expect(() => pickle.readBool('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = PickleBuilder().withString('key', 'true').build();

      // then
      expect(() => pickle.readBool('key'), throwsA(isA<PickleTypeError>()));
    });
  });

  group('withDouble', () {
    test('assigned and read', () {
      // given
      final pickle = PickleBuilder().withDouble('key', 2.4).build();

      // when
      final result = pickle.readDouble('key');

      // then
      expect(result, 2.4);
    });
    test('without assignment', () {
      // given
      final pickle = PickleBuilder().build();

      // then
      expect(() => pickle.readDouble('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = PickleBuilder().withString('key', '2.4').build();

      // then
      expect(() => pickle.readDouble('key'), throwsA(isA<PickleTypeError>()));
    });
  });

  group('withInt', () {
    test('assigned and read', () {
      // given
      final pickle = PickleBuilder().withInt('key', 2).build();

      // when
      final result = pickle.readInt('key');

      // then
      expect(result, 2);
    });
    test('without assignment', () {
      // given
      final pickle = PickleBuilder().build();

      // then
      expect(() => pickle.readInt('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = PickleBuilder().withString('key', '2').build();

      // then
      expect(() => pickle.readInt('key'), throwsA(isA<PickleTypeError>()));
    });
  });

  group('withString', () {
    test('assigned and read', () {
      // given
      final pickle = PickleBuilder().withString('key', 'SUCCESS').build();

      // when
      final result = pickle.readString('key');

      // then
      expect(result, 'SUCCESS');
    });
    test('without assignment', () {
      // given
      final pickle = PickleBuilder().build();

      // then
      expect(() => pickle.readString('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = PickleBuilder().withInt('key', 2).build();

      // then
      expect(() => pickle.readString('key'), throwsA(isA<PickleTypeError>()));
    });
  });

  group('withPickle', () {
    test('assigned and read', () {
      // given
      final pickle = PickleBuilder()
          .withPickle('key', PickleBuilder().withInt('lock', 2).build())
          .build();

      // when
      final result = pickle.readPickle('key');

      // then
      expect(result.readInt('lock'), 2);
    });
    test('without assignment', () {
      // given
      final pickle = PickleBuilder().build();

      // then
      expect(() => pickle.readPickle('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = PickleBuilder().withString('key', 'Pickle').build();

      // then
      expect(() => pickle.readPickle('key'), throwsA(isA<PickleTypeError>()));
    });
  });

  group('withPickles', () {
    test('assigned and read', () {
      // given
      final pickle =
          PickleBuilder().withPickles('key', [PickleBuilder().build()]).build();

      // when
      final result = pickle.readPickles('key');

      // then
      expect(result.length, 1);
    });
    test('without assignment', () {
      // given
      final pickle = PickleBuilder().build();

      // then
      expect(() => pickle.readPickles('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = PickleBuilder().withString('key', '[]').build();

      // then
      expect(() => pickle.readPickles('key'), throwsA(isA<PickleTypeError>()));
    });
  });

  group('withPickleable', () {
    test('assigned and read', () {
      // given
      final pickle =
          PickleBuilder().withPickleable('key', Tester(name: 'kiwi')).build();

      // when
      final result = pickle.readPickleable('key', Tester.fromPickle);

      // then
      expect(result.name, 'kiwi');
    });
    test('without assignment', () {
      // given
      final pickle = PickleBuilder().build();

      // then
      expect(() => pickle.readPickleable('key', Tester.fromPickle),
          throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = PickleBuilder().withString('key', 'Tester').build();

      // then
      expect(() => pickle.readPickleable('key', Tester.fromPickle),
          throwsA(isA<PickleTypeError>()));
    });
  });

  group('withPickleables', () {
    test('assigned and read', () {
      // given
      final pickle = PickleBuilder().withPickleables(
          'key', [Tester(name: 'Henry'), Tester(name: 'Henrietta')]).build();

      // when
      final result = pickle.readPickleables('key', Tester.fromPickle);

      // then
      expect(result, [Tester(name: 'Henry'), Tester(name: 'Henrietta')]);
    });
    test('without assignment', () {
      // given
      final pickle = PickleBuilder().build();

      // then
      expect(() => pickle.readPickleables('key', Tester.fromPickle),
          throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = PickleBuilder().withString('key', '2').build();

      // then
      expect(() => pickle.readPickleables('key', Tester.fromPickle),
          throwsA(isA<PickleTypeError>()));
    });
  });

  group('readString', () {
    test('read successfully', () {
      // given
      final value = 'value';
      final pickle = PickleBuilder().withString('key', value).build();

      // when
      final result = pickle.readString('key');

      // then
      expect(result, value);
    });
    test('read successfully with default value', () {
      // given
      final value = 'value';
      final pickle = PickleBuilder().withString('key', value).build();

      // when
      final result = pickle.readString('key', defaultValue: 'alternateValue');

      // then
      expect(result, value);
    });
    test('read without a value', () {
      // given
      final pickle = PickleBuilder().withString('key', 'value').build();

      // then
      expect(() => pickle.readString('wrongKey'), throwsArgumentError);
    });
    test('read without a value but a default', () {
      // given
      final value = 'alternateValue';
      final pickle = PickleBuilder().withString('key', 'value').build();

      // when
      final result = pickle.readString('wrongKey', defaultValue: value);

      // then
      expect(result, value);
    });
  });
  group('readBool', () {
    test('read successfully', () {
      // given
      final value = true;
      final pickle = PickleBuilder().withBool('key', value).build();

      // when
      final result = pickle.readBool('key');

      // then
      expect(result, value);
    });
    test('read successfully with default value', () {
      // given
      final value = true;
      final pickle = PickleBuilder().withBool('key', value).build();

      // when
      final result = pickle.readBool('key', defaultValue: false);

      // then
      expect(result, value);
    });
    test('read without a value', () {
      // given
      final pickle = PickleBuilder().withBool('key', true).build();

      // then
      expect(() => pickle.readBool('wrongKey'), throwsArgumentError);
    });
    test('read without a value but a default', () {
      // given
      final value = false;
      final pickle = PickleBuilder().withBool('key', true).build();

      // when
      final result = pickle.readBool('wrongKey', defaultValue: value);

      // then
      expect(result, value);
    });
  });
  group('readInt', () {
    test('read successfully', () {
      // given
      final value = 3;
      final pickle = PickleBuilder().withInt('key', value).build();

      // when
      final result = pickle.readInt('key');

      // then
      expect(result, value);
    });
    test('read successfully with default value', () {
      // given
      final value = 3;
      final pickle = PickleBuilder().withInt('key', value).build();

      // when
      final result = pickle.readInt('key', defaultValue: 5);

      // then
      expect(result, value);
    });
    test('read without a value', () {
      // given
      final pickle = PickleBuilder().withInt('key', 3).build();

      // then
      expect(() => pickle.readInt('wrongKey'), throwsArgumentError);
    });
    test('read without a value but a default', () {
      // given
      final value = 5;
      final pickle = PickleBuilder().withInt('key', 3).build();

      // when
      final result = pickle.readInt('wrongKey', defaultValue: value);

      // then
      expect(result, value);
    });
  });
  group('readDouble', () {
    test('read successfully', () {
      // given
      final value = 3.0;
      final pickle = PickleBuilder().withDouble('key', value).build();

      // when
      final result = pickle.readDouble('key');

      // then
      expect(result, value);
    });
    test('read successfully with default value', () {
      // given
      final value = 3.0;
      final pickle = PickleBuilder().withDouble('key', value).build();

      // when
      final result = pickle.readDouble('key', defaultValue: 5.0);

      // then
      expect(result, value);
    });
    test('read without a value', () {
      // given
      final pickle = PickleBuilder().withDouble('key', 3.0).build();

      // then
      expect(() => pickle.readDouble('wrongKey'), throwsArgumentError);
    });
    test('read without a value but a default', () {
      // given
      final value = 5.0;
      final pickle = PickleBuilder().withDouble('key', 3).build();

      // when
      final result = pickle.readDouble('wrongKey', defaultValue: value);

      // then
      expect(result, value);
    });
  });
  group('readPickle', () {
    test('read successfully', () {
      // given
      final value = PickleBuilder().withString('key', 'value').build();
      final pickle = PickleBuilder().withPickle('key', value).build();

      // when
      final result = pickle.readPickle('key');

      // then
      expect(result, value);
    });
    test('read successfully with default value', () {
      // given
      final value = PickleBuilder().withString('key', 'value').build();
      final pickle = PickleBuilder().withPickle('key', value).build();

      // when
      final result =
          pickle.readPickle('key', defaultValue: PickleBuilder().build());

      // then
      expect(result, value);
    });
    test('read without a value', () {
      // given
      final pickle =
          PickleBuilder().withPickle('key', PickleBuilder().build()).build();

      // then
      expect(() => pickle.readDouble('wrongKey'), throwsArgumentError);
    });
    test('read without a value but a default', () {
      // given
      final value = PickleBuilder().withString('key', 'value').build();
      final pickle =
          PickleBuilder().withPickle('key', PickleBuilder().build()).build();

      // when
      final result = pickle.readPickle('wrongKey', defaultValue: value);

      // then
      expect(result, value);
    });
  });
  group('readPickleable', () {
    test('read successfully', () {
      // given
      final value = Tester();
      final pickle = PickleBuilder().withPickleable('key', value).build();

      // when
      final result = pickle.readPickleable('key', Tester.fromPickle);

      // then
      expect(result, value);
    });
    test('read successfully with default value', () {
      // given
      final value = Tester();
      final pickle = PickleBuilder().withPickleable('key', value).build();

      // when
      final result = pickle.readPickleable('key', Tester.fromPickle,
          defaultValue: Tester(name: 'Ernest'));

      // then
      expect(result, value);
    });
    test('read without a value', () {
      // given
      final pickle = PickleBuilder().withPickleable('key', Tester()).build();

      // then
      expect(() => pickle.readPickleable('wrongKey', Tester.fromPickle),
          throwsArgumentError);
    });
    test('read without a value but a default', () {
      // given
      final value = Tester();
      final pickle =
          PickleBuilder().withPickleable('key', Tester(name: 'Ernest')).build();

      // when
      final result = pickle.readPickleable('wrongKey', Tester.fromPickle,
          defaultValue: value);

      // then
      expect(result, value);
    });
  });
  group('readPickles', () {
    test('read successfully', () {
      // given
      final value = [
        PickleBuilder().withString('key', 'value').build(),
        PickleBuilder().build()
      ];
      final pickle = PickleBuilder().withPickles('key', value).build();

      // when
      final result = pickle.readPickles('key');

      // then
      expect(result, value);
    });
    test('read successfully with default value', () {
      // given
      final value = [
        PickleBuilder().withString('key', 'value').build(),
        PickleBuilder().build()
      ];
      final pickle = PickleBuilder().withPickles('key', value).build();

      // when
      final result = pickle.readPickles('key', defaultValue: []);

      // then
      expect(result, value);
    });
    test('read without a value', () {
      // given
      final pickle = PickleBuilder().withPickles('key', []).build();

      // then
      expect(() => pickle.readPickles('wrongKey'), throwsArgumentError);
    });
    test('read without a value but a default', () {
      // given
      final value = [PickleBuilder().build()];
      final pickle = PickleBuilder().withPickles('key', []).build();

      // when
      final result = pickle.readPickles('wrongKey', defaultValue: value);

      // then
      expect(result, value);
    });
  });
  group('readPickleables', () {
    test('read successfully', () {
      // given
      final value = [Tester(name: 'John')];
      final pickle = PickleBuilder().withPickleables('key', value).build();

      // when
      final result = pickle.readPickleables('key', Tester.fromPickle);

      // then
      expect(result, value);
    });
    test('read successfully with default value', () {
      // given
      final value = [Tester(name: 'John')];
      final pickle = PickleBuilder().withPickleables('key', value).build();

      // when
      final result = pickle.readPickleables('key', Tester.fromPickle,
          defaultValue: [Tester(name: 'Ernest')]);

      // then
      expect(result, value);
    });
    test('read without a value', () {
      // given
      final pickle = PickleBuilder().withPickleables('key', [Tester()]).build();

      // then
      expect(() => pickle.readPickleables('wrongKey', Tester.fromPickle),
          throwsArgumentError);
    });
    test('read without a value but a default', () {
      // given
      final value = [Tester(name: 'Jimmy')];
      final pickle = PickleBuilder()
          .withPickleables('key', [Tester(name: 'Ernest')]).build();

      // when
      final result = pickle.readPickleables('wrongKey', Tester.fromPickle,
          defaultValue: value);

      // then
      expect(result, value);
    });
  });
}
